;; Generated package description from exwm.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "exwm" "0.35.0.20260704.2" "Emacs X Window Manager" '((emacs "29.1") (xelb "0.23") (compat "31")) :commit "2d0fb41793f8bc50abdea394f48b4653e8317286" :authors '(("Chris Feng" . "chris.w.feng@gmail.com")) :maintainer '(("Adrián Medraño Calvo" . "adrian@medranocalvo.com") ("Steven Allen" . "steven@stebalien.com") ("Daniel Mendler" . "mail@daniel-mendler.de")) :keywords '("unix") :url "https://github.com/emacs-exwm/exwm")
